@@ -27,7 +27,7 @@ def prune_incomplete(g):
     for k in g.nodes():
         gk = g[k]
         for l in g.neighbors(k):
-            if k < l:
+            if k < l and gk[l]['type'] < 0:
                 gl = g[l]
                 for m in g.neighbors(l):
                     if g.has_edge(k,m) and gk[l]['weight'] > gk[m]['weight']+gl[m]['weight']:
@@ -41,8 +41,10 @@ def prune_complete(g):
     to_prune = set()
     for k in g.nodes():
         for l in g.neighbors(k):
-            if k < l and g[k][l]['weight'] > astar_path_length(g,k,l):
-                to_prune.add((k,l))
+            if k < l:
+                gkl = g[k][l]
+                if gkl['type'] < 0 and gkl['weight'] > astar_path_length(g,k,l):
+                    to_prune.add((k,l))
     for u,v in to_prune:
         if g.has_edge(u,v):
             g.remove_edge(u,v)
@@ -107,9 +109,6 @@ def extract_cache(file_name_in,file_name_out,overpass_url,area=None,around=1000,
         replace(file_name_tmp,file_name_in)
         end('')
         file_size(file_name_in)
-    common_end(g,file_name_out)
-        
-def common_end(g,file_name_out,id2types=None):
     from limic.util import start, end, status, file_size, save_pickled
     from networkx import relabel_nodes
     start("Prune redundant edges (incomplete)")
@@ -123,122 +122,17 @@ def common_end(g,file_name_out,id2types=None):
     start("Cleaning up graph")
     relabel = dict(map(lambda tower:(tower,(tower.id,tower.latlon[0],tower.latlon[1])),g.nodes()))
     relabel_nodes(g,relabel,copy=False)
-    if id2types:
-        for u,v,d in g.edges(data=True):
-            if u[0] >= 0:
-                current = u
-            elif v[0] >= 0:
-                current = v
-            else:
-                todo = [u]
-                done = set()
-                current = None
-                while todo:
-                    current = todo.pop()
-                    if current[0] >= 0:
-                        break
-                    if current in done:
-                        continue
-                    done.add(current)
-                    for n in g.neighbors(current):
-                        if not n in done:
-                            todo.append(n)
-                assert(current is not None)
-            d['type'] = -1 if d['type'] else id2types[current[0]]
-    else:
-        for u,v,d in g.edges(data=True):
-            d['type'] = -1 if d['type'] else 0
+    for u,v,d in g.edges(data=True):
+        d['type'] = -1 if d['type'] else 0
     end()
     start("Saving graph to",file_name_out)
     save_pickled(file_name_out,g)
     end('')
     file_size(file_name_out)
 
-def is_safe(m_nodes,id2substations,id2elem,tower,around):
-    from shapely.geometry import Point, Polygon
-    nodes = [] if m_nodes[1] is None else list(map(lambda x:m_nodes[0][x],m_nodes[1].query_ball_point(m_nodes[2].transform(*tower.latlon),r=around)))
-    if not nodes:
-        return True
-    p = Point(*tower.latlon)
-    substations = []
-    for node in nodes:
-        for substation in id2substations[node.id]:
-            if not substation in substations:
-                substations.append(substation)
-    for substation in substations:
-        if len(substation) < 3:
-            continue
-        poly = Polygon(list(map(lambda x:id2elem[x].latlon,substation)))
-        if p.within(poly):
-            return False
-    return True
-
-def find_neighbours(tower,id2elem,id2lines):
-    towers = set()
-    if tower.id < 0:
-        towers.update(tower.neighbours)
-        return towers
-    for line in id2lines[tower.id]:
-        index = line.index(tower.id)
-        if index > 0:
-            towers.add(line[index-1])
-        if index+1 < len(line):
-            towers.add(line[index+1])
-    return list(map(lambda id:id2elem[id],towers))
-
-def find_segments(tower,around,id2elem,id2lines,neighbourhood):
-    segments = []
-    lines = []
-    for other in neighbourhood:
-        for line in id2lines[other.id]:
-            if not line in lines:
-                lines.append(line)
-    for line in lines:
-        for i in range(1,len(line)):
-            segments.append((line[i-1],line[i]))
-    return list(map(lambda x:(id2elem[x[0]],id2elem[x[1]]),segments))            
-
-def find_all_neighbours(tower,around,eps,safe_dist,minusid,latlon2id,penalize,m_towers,m_nodes,id2elem,id2lines,id2substations,latlon2safe):
-    from limic.overpass import dist, dist_towers, intersect, pylon
-    from limic.util import distance
-    direct_neighbours = find_neighbours(tower,id2elem,id2lines)
-    neighbours = dist_towers(direct_neighbours,tower.latlon)
-    direct_segments = list(map(lambda neighbour:(neighbour.tower,tower),neighbours))
-    #print(tower)
-    #print(direct_neighbours)
-    #print(neighbours)
-    neighbourhood = [] if m_towers[1] is None else list(map(lambda x:m_towers[0][x],m_towers[1].query_ball_point(m_towers[2].transform(*tower.latlon),r=around)))
-    for neighbour in map(lambda x:dist(0.,False,x),neighbourhood):
-        if not neighbour.tower in direct_neighbours and neighbour.tower.id != tower.id:
-            neighbour.dist = distance(neighbour.tower.latlon,tower.latlon)*penalize
-            neighbour.air = True
-            neighbours.append(neighbour)
-            #print(neighbour)
-    indirect_segments = find_segments(tower,around,id2elem,id2lines,neighbourhood)
-    for indirect_segment in indirect_segments:
-        for direct_segment in direct_segments:
-            intersection = intersect(direct_segment[0].latlon,direct_segment[1].latlon,indirect_segment[0].latlon,indirect_segment[1].latlon,eps)
-            if intersection:
-                #if verbosity >= 3:
-                #    print("found intersection",direct_segment,indirect_segment,intersection)
-                #    print("node(id:"+(",".join(map(lambda t:str(t.id),[direct_segment[0],direct_segment[1],indirect_segment[0],indirect_segment[1]])))+");out;")
-                if not intersection in latlon2id:
-                    minusid[0] -= 1
-                    latlon2id[intersection] = minusid[0]
-                neighbours.append(dist(distance(intersection,tower.latlon),False,pylon(latlon2id[intersection], intersection,neighbours=indirect_segment)))
-    safe_neighbours = []
-    for neighbour in neighbours:
-        safe = latlon2safe.get(neighbour.tower.latlon,None)
-        if safe == None:
-            safe = is_safe(m_nodes,id2substations,id2elem,neighbour.tower,safe_dist)
-            latlon2safe[neighbour.tower.latlon] = safe
-        if safe:
-            safe_neighbours.append(neighbour)
-        #else:
-            #if verbosity >= 4: print("unsafe and skipped:",neighbour.tower)
-    return safe_neighbours
-
 def extract_osm(file_name_in,file_name_out,around=1000,eps=0.01,safe_dist=100,penalize=20,white="{'power':'line'}",black="{'power':'substation'}",conserve_mem=False):
+    from setproctitle import setproctitle
+    setproctitle(file_name_in)
     from limic.util import check_overwrite
     if not check_overwrite(file_name_in,file_name_out):
         return
@@ -268,60 +162,129 @@ def extract_osm_post(file_name_in,file_name_out,around=1000,eps=0.01,safe_dist=1
     osm_post(lim,file_name_out,around=1000,eps=0.01,safe_dist=100,penalize=20)
 
 def osm_post(lim,file_name_out,around=1000,eps=0.01,safe_dist=100,penalize=20):
-    from limic.util import start, end, status, file_size, load_pickled, haversine_distance
-    #from limic.mtree import MTree
+    from limic.util import start, end, status, file_size, load_pickled, distance, save_pickled
     from scipy.spatial import cKDTree as KDTree
-    from networkx import Graph
+    from networkx import Graph, astar_path_length
     from pyproj import CRS, Transformer
-    lines, substations, towers, nodes, id2elem, id2lines, id2substations, id2types = lim
-    #start("Loading filtered OSM data from",file_name_in)
-    #lines, substations, towers, nodes, id2elem, id2lines, id2substations = load_pickled(file_name_in)
-    #end()
-    start("Building KD-trees from nodes")
-    #def distance(x,y):
-    #    return haversine_distance(longx=x.latlon[1],latx=x.latlon[0],longy=y.latlon[1],laty=y.latlon[0])
-    #m_towers = MTree(d=distance)
-    #m_towers.add_all(towers)
-    #m_nodes = MTree(d=distance)
-    #m_nodes.add_all(nodes)
-    crs_4326 = CRS("WGS84")
-    crs_proj = CRS("EPSG:28992")
-    transformer = Transformer.from_crs(crs_4326, crs_proj)
-    towers_tree = KDTree(list(map(lambda x:transformer.transform(x.latlon[0],x.latlon[1]),towers))) if towers else None
-    m_towers = (towers,towers_tree,transformer)
-    nodes_tree = KDTree(list(map(lambda x:transformer.transform(x.latlon[0],x.latlon[1]),nodes))) if nodes else None
-    m_nodes = (nodes,nodes_tree,transformer)
-    end()
-    #print(m_towers.size)
-    #print(m_nodes.size)
-    #print("node(id:"+(",".join(list(map(lambda x:str(x.obj.id) if x else "",m_towers.search(towers[0],r=3000)))))+");out;")
-    from limic.overpass import pylon
-    #find_segments(pylon(1847250054,(56.3973359,8.7254833)),around,m_towers,id2elem,id2lines)
-    start("Building safe nodes")
-    latlon2safe = {}
-    g=Graph()
-    for tower in towers:
-        safe = is_safe(m_nodes,id2substations,id2elem,tower,safe_dist)
-        latlon2safe[tower.latlon] = safe
-        if safe:
-            g.add_node(tower)
-#        else:
-#            if verbosity >= 2: print("NOT safe!")
+    from itertools import chain
+    from limic.overpass import intersect, pylon
+    lines, substations, towers, id2tower, id2node, id2lines, id2types = lim
+    start("Building KD-tree from white nodes")
+    from limic.util import kdtree
+    towers_tree = kdtree(towers,get_latlon=lambda x:x.latlon)
     end('')
-    total = len(g.nodes())
-    status(total)
-    start("Building edges")
-    def fan(tower,around,eps,safe_dist,minusid,latlon2id,penalize):
-        return find_all_neighbours(tower,around,eps,safe_dist,minusid,latlon2id,penalize,m_towers,m_nodes,id2elem,id2lines,id2substations,latlon2safe)
-    build_edges(g,fan,around,eps,safe_dist,penalize)
+    status(len(towers))
+    start("Deleting black nodes")
+    to_delete = set()
+    from limic.util import nodes_in_geometry
+    for substation in substations:
+        to_delete.update(nodes_in_geometry(towers_tree,list(map(lambda x:id2node[x],substation))))
+    towers = [tower for tower in towers if tower not in to_delete]
+    end('')
+    status(len(towers))
+    start("Building initial graph")
+    g = Graph()
+    g.add_nodes_from(towers)
+    for line in lines:
+        line_nodes = list(map(lambda x:id2tower[x],line))
+        for from_node, to_node in zip(line_nodes,line_nodes[1:]):
+            if from_node in to_delete or to_node in to_delete:
+                continue
+            w = distance(from_node.latlon,to_node.latlon)
+            g.add_edge(from_node,to_node,weight=w,type=id2types[from_node.id])
+    end('')
+    status(len(g.nodes()),end='/')
+    status(len(g.edges()))
+    start("Finding neighbours within "+str(around)+"m")
+    towers_tree = kdtree(towers,get_latlon=lambda x:x.latlon)
+    end('')
+    neighbour_indices, neighbours = towers_tree.get_neighbours(around=1000)
+    end()
+    start("Computing non-logical intersections")
+    tower2index = {}
+    for i,t in zip(range(len(towers)),towers):
+        tower2index[t] = i
+    for k,v in id2lines.items():
+        id2lines[k] = tuple(map(tuple,v))
+    end('')
+    segments = set()
+    for u,v in g.edges():
+        this = (u,v) if u < v else (v,u)
+        ui, vi = tower2index[u], tower2index[v]
+        lines = set()
+        lines.update(id2lines[u.id])
+        lines.update(id2lines[v.id])
+        for neighbour in chain(neighbours[ui],neighbours[vi]):
+            if neighbour == u or neighbour == v:
+                continue
+            if not lines.intersection(id2lines[neighbour.id]):
+                for nn in g.neighbors(neighbour):
+                    other = (neighbour,nn) if neighbour < nn else (nn,neighbour)
+                    segments.add(tuple(sorted((this,other))))
+    end('')
+    status(len(segments),end='   ')
+    neighbours2intersection = {}
+    minusid = 0
+    latlon2id = {}
+    segments2intersections = {}
+    for (t1,t2),(t3,t4) in segments:
+        res = intersect(t1.latlon,t2.latlon,t3.latlon,t4.latlon,eps=eps,no_tu=False)
+        if res:
+            intersection,(t,u) = res
+            if not intersection in latlon2id:
+                minusid -= 1
+                latlon2id[intersection] = minusid
+            segments2intersections.setdefault((t1,t2),[]).append((t,latlon2id[intersection],intersection))
+            segments2intersections.setdefault((t3,t4),[]).append((u,latlon2id[intersection],intersection))
+    end('')
+    status(-minusid,end='   ')
+    for (u,v), intersections in segments2intersections.items():
+        intersections.sort()
+        g.remove_edge(u,v)
+        type = id2types[u.id]
+        assert(type == id2types[v.id])
+        seq = [u]
+        for _,id,latlon in intersections:
+            seq.append(pylon(id,latlon))
+        seq.append(v)
+        for from_node, to_node in zip(seq,seq[1:]):
+            w = distance(from_node.latlon,to_node.latlon)
+            g.add_edge(from_node,to_node,weight=w,type=type)
+    end()
+    start("Adding routing through air")
+    airs = set()
+    for ns in neighbours:
+        n = ns[0]
+        for m in ns[1:]:
+            if not g.has_edge(n,m):
+                airs.add((n,m))
+    end('')
+    for n,m in airs:
+        w = penalize*distance(n.latlon,m.latlon)
+        g.add_edge(n,m,weight=w,type=-1)
+    end('')
+    status(len(g.nodes()),end='/')
+    status(len(g.edges()))
+    from networkx import relabel_nodes
+    start("Prune redundant edges (incomplete)")
+    prune_incomplete(g)
     end('')
     status(len(g.edges()))
-    common_end(g,file_name_out,id2types)
+    start("Prune redundant edges (complete)")
+    prune_complete(g)
+    end('')
+    status(len(g.edges()))
+    start("Cleaning up graph")
+    relabel = dict(map(lambda tower:(tower,(tower.id,tower.latlon[0],tower.latlon[1])),g.nodes()))
+    relabel_nodes(g,relabel,copy=False)
+    end()
+    start("Saving graph to",file_name_out)
+    save_pickled(file_name_out,g)
+    end('')
+    file_size(file_name_out)
 
 def osm_pre(file_name_in,white=[("power","line")],black=[("power","substation")],conserve_mem=False):
-    from limic.util import start, end, status, file_size, save_pickled, split, file_copy
-    from xml.sax import parse
-    from xml.sax.handler import ContentHandler
+    from limic.util import start, end, status, file_size, split, file_copy
     from bz2 import open as bopen
     from limic.overpass import pylon
     white = list(map(lambda x:(x[0],x[1][0]+'" v="'+x[1][1]+'"'),zip(range(len(white)),white)))
@@ -357,6 +320,12 @@ def osm_pre(file_name_in,white=[("power","line")],black=[("power","substation")]
     id2types = {}
     for elem in elems:
         if elem.startswith('way id="'):
+            found = False
+            for i,t in white_black:
+                if t in elem:
+                    found = True
+            if not found:
+                continue
             tags = elem.split('\n    <tag k="')
             for tag in tags[1:]:
                 for i,t in white_black:
@@ -386,29 +355,29 @@ def osm_pre(file_name_in,white=[("power","line")],black=[("power","substation")]
         area.seek(0)
         elems = split(area,'\n  <',maxlength=area_length)
     towers = []
-    nodes = []
-    id2elem = {}
+    id2tower = {}
+    id2node = {}
+    ids = set()
+    ids.update(map(lambda x:str(x),id2lines.keys()))
+    ids.update(map(lambda x:str(x),id2substations.keys()))
     for elem in elems:
         if elem.startswith('node id="'):
+            if not elem[9:elem.find('"',10)] in ids:
+                continue
             parts = elem.split("\n")[0].split('"')
-            if parts[-5]!=' lat=':
-                print(parts)
-            if parts[-3]!=' lon=':
-                print(parts)
+            assert(parts[-5]==' lat=')
+            assert(parts[-3]==' lon=')
             i = int(parts[1])
             if i in id2lines:
                 elem = pylon(i,(float(parts[-4]),float(parts[-2])))
                 towers.append(elem)
-                id2elem[i] = elem
+                id2tower[i] = elem
             if i in id2substations:
-                elem = pylon(i,(float(parts[-4]),float(parts[-2])))
-                nodes.append(elem)
-                id2elem[i] = elem
+                elem = (float(parts[-4]),float(parts[-2]))
+                id2node[i] = elem
     end('')
-    status(len(towers),end='   ')
-    status(len(nodes))
-    start("Checking that all nodes were extracted for",file_name_in)
+    status(len(towers))
+    start("Checking all nodes were extracted")
     assert(set(id2lines.keys())==set(map(lambda x:x.id,towers)))
-    assert(set(id2substations.keys())==set(map(lambda x:x.id,nodes)))
     end()
-    return lines,substations,towers,nodes,id2elem,id2lines,id2substations,id2types
+    return lines, substations, towers, id2tower, id2node, id2lines, id2types
